# $Id: Output.py 659 2007-11-05 21:05:51Z indraneel $

import os
from Bio.PDB.PDBIO import PDBIO
import gzip

pdbio = PDBIO()

class PSSDFileNames:
    def __init__(self, conf):
        self.conf = conf

    def get_outfname(self):
        conf = self.conf
        secoutdir = conf.get('Output','secoutdir')
        structureID = conf.get('Input', 'structureID')
        endtag = conf.get('Output', 'secfilename_posttag')
        outfname = secoutdir+structureID+endtag
        return outfname

    def getVecFileName(self):
        conf = self.conf
        vecoutdir = conf.get('Output','vectoroutdir')
        structureID = conf.get('Input', 'structureID')
        endtag = conf.get('Output', 'vecfilename_posttag')
        vecoutfname = vecoutdir+structureID+endtag
        return vecoutfname

class PSSDPDBOut(PSSDFileNames):
    def __init__(self, structure, header, trailer, conf, logger):
        self.conf = conf
        PSSDFileNames.__init__(self, conf)
        self.outfname = self.get_outfname()
        self.vectorFileName = self.getVecFileName()
        self.structure = structure
        self.header = header
        self.trailer = trailer
        self.logger = logger

    def _get_definition_lines(self, definition):
        lines = []
        helix_ctr = 0
        definition.sort_by_position()
        for type in ['H', 'E', 'S', 'L']:
            for elem in definition:
                if elem.type != type: continue
                if elem.type == 'H':
                    helix_ctr += 1
                    lines += [elem.get_H_line(helix_ctr)]
                elif elem.type == 'E':
                    lines += [elem.get_E_line()]
                elif elem.type == 'S':
                    lines += [elem.get_S_line()]
                elif elem.type == 'L':
                    lines += [elem.get_L_line()]
        if lines == []:
            self.logger.log('info', 'No secondary structure found')
        return lines

    def _rescmp(self, resa, resb):
        return cmp(resa.strpos, resb.strpos)

    def _get_pair_lines(self, pairs):
        lines = []
        pair_ctr = 0
        keylist = pairs.keys()
        keylist.sort(self._rescmp)
        for res in keylist:
            for pair in pairs[res]:
                pair_ctr += 1
                line = "PAIRS "+" "+"   "+" "+"   "+" "
                line += res.get_resname() + " "
                line += res.get_parent().get_id() + " "
                line += "%4d" % (res.get_id()[1]) + res.get_id()[2] + " "
                line += pair.get_resname() + " "
                line += pair.get_parent().get_id() + " "
                line += "%4d" % pair.get_id()[1]
                line += pair.get_id()[2]
                line += '\n'
                lines += [line]
        return lines

    def _replaceHeader(self, lines, header):
        # remove empty lines from original header
        header = filter(lambda line: line[:-1].strip()!='', header)
        
        sidx = 0
        for (sidx,line) in enumerate(header):
            if line[0:6] == "HELIX " or \
               line[0:6] == "SHEET ":
                break
        if sidx:
            eidx = 0
            for line in header:
                if line[0:6] == "HELIX " or \
                       line[0:6] == "SHEET ":
                    eidx = header.index(line)
        if sidx:
            header = header[:sidx] + header[eidx+1:]
        else:
            for (sidx,line) in enumerate(header):
                if line[0:6] == "TURN  " or \
                       line[0:6] == "SSBOND" or \
                       line[0:6] == "LINK  ":
                    break
        if sidx:
            header = header[:sidx] + lines + header[sidx:]
        else:
            header = header + lines
        return header

    def _write(self, lines):
        header = self.header
        trailer = self.trailer
        
        header = self._replaceHeader(lines, header)
        compression = self.conf.getint('Output', 'compression')
        if compression:
            outfh = gzip.GzipFile(self.outfname+'.gz', 'w')
        else:
            outfh = file(self.outfname, 'w')
        outfh.writelines(header)
        pdbio.set_structure(self.structure)
        pdbio.save(outfh)
        outfh.writelines(trailer)
        outfh.close()

    def _writeVector(self, definition):
        definition.sort_by_position()
        lines = []
        for elem in definition:
            if elem.type not in ['H', 'E', 'S', 'L']: continue
            (vst, ven) = elem.getVectorEnds()
            vst = map(lambda x:'%8.3f'%x, vst)
            ven = map(lambda x:'%8.3f'%x, ven)
            chain = elem[0].get_parent().get_id()
            (stid, enid) = (elem[0].get_id()[1:], elem[-1].get_id()[1:])
            (stn, enn) = (elem[0].get_resname(), elem[-1].get_resname())
            vals = (elem.type, tuple(vst), tuple(ven))
            vals += (chain, stn, stid, enn, enid)
            lines += [str(vals)+'\n']
        compression = self.conf.getint('Output', 'compression')
        if compression:
            outfh = gzip.GzipFile(self.vectorFileName+'.gz', 'w')
        else:
            outfh = file(self.vectorFileName, 'w')
        outfh.writelines(lines)
        outfh.close()        

    # Public Methods

    def write(self, definition):
        lines = self._get_definition_lines(definition)
        pairs = definition.get_E_pairs()
        lines += self._get_pair_lines(pairs)
        self._write(lines)

    def writeVector(self, definition):
        self._writeVector(definition)
