# $Id: PSSDParser.py 577 2006-10-11 16:51:51Z indraneel $

# basic PSSD output parser
# stores in Chain Residue format and also as elements

import os
import sys
import string
import traceback

from Log import LogFH
from Bio.PDB.PDBParser import PDBParser
from Bio.SCOP.Raf import to_one_letter_code
from Element import PSSDElement
from Definition import PSSDefinition

class Element:
    def __repr__(self):
        line = "< %c %c %4d %c %4d %c >" % (self.chain, self.type, self.start, self.starticode, self.end, self.endicode)
        return line

class PSSDResidue:
    def __init__(self, logger, BioPDBRes, sstype):
        self.logger = logger
        self.BioPDBRes = BioPDBRes

        res = self.BioPDBRes
        try:
            self.serial=None
            self.resid=res.id[1]
            self.icode=res.id[2]
            self.chain=res.get_parent().id
            self.aa=to_one_letter_code[res.resname]
            self.structure=sstype
            self.type=self.structure
            self.BP1=None
            self.BP2=None
            (x,y,z)=res['CA'].coord
            self.coord = res['CA'].coord
        except:
            logger = self.logger
            [etype, val, tb] = sys.exc_info()
            errors = traceback.format_exception(etype, val, tb)
            logger.log(errors)

    def __getattr__(self, name):
        return getattr(self.BioPDBRes, name)

    def __repr__(self):
        line=''
        line+=self.chain+' '+str(self.resid)+self.icode+' '+self.structure
        for attr in ['pair1', 'pair2']:
            if hasattr(self, attr):
                line+=' '+ '%4d' % (getattr(self,attr)).resid
            else: line+='     '
        line+=' '+self.type
        line+=' '+str(self.coord)
        return line

    def pairBP(self):
        pass

class ParsedDefinition(dict):
    def __init__(self, logger=None, name=''):
        self.logger = logger
        self.name = name

    def mkDefinition(self, structure, oldelements):
        definition = PSSDefinition(self.logger, name=self.name)
        for oldelement in oldelements:
            element = PSSDElement(self.logger, oldelement.type)
            definition.append(element)

        model = structure.get_list()[0]
        ctr = 0
        buf_elemlist = [False] * len(oldelements)
        for chain in model.get_list():
            self[chain.id] = {}
            structure = {'H':0, 'E':0} # warning
            for res in chain.get_list():
                if not to_one_letter_code.has_key(res.resname):
                    continue
                ctr += 1
                aa = to_one_letter_code[res.resname]
                self[chain.id][(res.id[1], res.id[2])] = \
                                           {'#':ctr, 'AA':aa, 'SS':[]}
                residue = self[chain.id][(res.id[1], res.id[2])]
                for (i, elem) in enumerate(oldelements):
                    if ((chain.id == elem.chain) and \
                        (res.id[1] == elem.start ) and \
                        (res.id[2] == elem.starticode)):
                        structure[elem.type] += 1
                        buf_elemlist[i] = True
                        
                if structure['E']:
                    residue['SS'].append('E')
                if structure['H']:
                    residue['SS'].append('H')
                if not residue['SS']:
                    residue['SS'].append(' ')

                for (i, val) in enumerate(buf_elemlist):
                    if not val: continue
                    pssd_res = PSSDResidue(self.logger, \
                                           res, definition[i].type)
                    definition[i].append(pssd_res)

                for (i, elem) in enumerate(oldelements):
                    if ((chain.id == elem.chain) and \
                        (res.id[1] == elem.end ) and \
                        (res.id[2] == elem.endicode)):
                        structure[elem.type] -= 1
                        buf_elemlist[i] = False
        self.definition = definition

    def getDefinition(self):
        if hasattr(self, 'definition'):
            return self.definition
        else: return 'Err'

class PSSD(ParsedDefinition):
    def __init__(self, logger=None):
        if logger:
            self.logger = logger
            
    def __sort(self, a, b):
        return cmp(self[self.chainID][a]['#'], self[self.chainID][b]['#'])

    def _getHeaderElements(self, lines):
        buf=[]
        for line in lines:
            if ( (line[:5] == 'HELIX') or
                 (line[:5] == 'SHEET') ):
                buf.append(line)
        sslines = buf
        oldelements = []
        for line in sslines:
            element = Element()
            if line[:5] == 'HELIX':
                element.chain = line[19]
                element.type = 'H'
                element.start = int(line[21:25])
                element.starticode = line[25]
                element.end = int(line[33:37])
                element.endicode = line[37]
            else:
                element.chain = line[21]
                element.type = 'E'
                element.start = int(line[22:26])
                element.starticode = line[26]
                element.end = int(line[33:37])
                element.endicode = line[37]
            oldelements.append(element)
        return oldelements

    def _getStructure(self, fname):
        parser = PDBParser(PERMISSIVE=1)
        try:
            structure = parser.get_structure('xxx', fname)
        except:
            self.logger.log('Bio.PDB could not parse: '+fname)
            return 'Err'
        return structure

    def store(self, fname):
        if not os.path.isfile(fname): return 'Err'
        self.fname = os.path.basename(fname)
        lines=file(fname).readlines()
        oldelements = self._getHeaderElements(lines)

        structure = self._getStructure(fname)
        if structure == 'Err': return 'Err'
        self.structure = structure # backward compat

        self.mkDefinition(structure, oldelements)
        self.elements = self.definition # backward compat

    def show(self, outfh):
        self.chainID = self.keys()[0]
	residues = self[self.chainID].keys()
	residues.sort(self.__sort)
	#for residue in residues:
	#    print >> outfh, self[self.chainID][residue]['SS'][0]
        print self.definition
        #self.definition.log()

if __name__ == '__main__':
    logger = LogFH('x')
    pssd = PSSD(logger)
    pssd.store(sys.argv[1])
    pssd.show(sys.stdout)
